"""
The `webfriend` module provides access to the browser automation and debugging facilities of
Google Chrome using the built-in [Chrome DevTools](https://developer.chrome.com/devtools) interface.
"""
# flake8: noqa
from __future__ import absolute_import
from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)  # noqa
